﻿using SocketTest.Helpers;
using System;
using System.Net;
using System.Net.Sockets;

namespace SocketTest.Sockets
{
    class Client : IDisposable
    {

        public event Action Connected;
        public event Action Disconnected;

        Socket client;

        public Client() { }

        void initSocket()
        {
            UninitSocket();

            client = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            client.LingerState = new LingerOption(true, 0);
            client.UseOnlyOverlappedIO = true;
            SocketHelper.SetKeepAlive(client);
        }

        public void Connect(IPEndPoint _ipEP)
        {
            initSocket();

            var connect_args = new SocketAsyncEventArgs();
            connect_args.RemoteEndPoint = _ipEP;
            connect_args.Completed += Connect_Completed;

            client.ConnectAsync(connect_args);
        }

        private void Connect_Completed(object sender, SocketAsyncEventArgs e)
        {

            var _socket = sender as Socket;

            try
            {
                if (e.SocketError == SocketError.Success)
                {
                    var recvEv = new SocketAsyncEventArgs();
                    var buf = new byte[1];
                    recvEv.SetBuffer(buf, 0, buf.Length);
                    recvEv.Completed += Receive_Completed;

                    _socket.ReceiveAsync(recvEv);

                    Connected?.Invoke();
                }
            }
            catch (ObjectDisposedException ex)
            {
                Disconnected?.Invoke();
            }
        }

        private void Receive_Completed(object sender, SocketAsyncEventArgs e)
        {
            var sock = sender as Socket;

            if (e.SocketError == SocketError.Success && sock.IsConnected() && e.BytesTransferred == 1)
            {
                sock.ReceiveAsync(e);
            }
        }

        void UninitSocket()
        {
            client?.Close();
            client?.Dispose();
            client = null;
        }

        #region IDisposable Support

        private bool disposedValue;

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    UninitSocket();
                    // TODO: 관리형 상태(관리형 개체)를 삭제합니다.
                }

                // TODO: 비관리형 리소스(비관리형 개체)를 해제하고 종료자를 재정의합니다.
                // TODO: 큰 필드를 null로 설정합니다.
                disposedValue = true;
            }
        }

        // // TODO: 비관리형 리소스를 해제하는 코드가 'Dispose(bool disposing)'에 포함된 경우에만 종료자를 재정의합니다.
        // ~Client()
        // {
        //     // 이 코드를 변경하지 마세요. 'Dispose(bool disposing)' 메서드에 정리 코드를 입력합니다.
        //     Dispose(disposing: false);
        // }

        public void Dispose()
        {
            // 이 코드를 변경하지 마세요. 'Dispose(bool disposing)' 메서드에 정리 코드를 입력합니다.
            Dispose(disposing: true);
            GC.SuppressFinalize(this);
        }

        #endregion

    }
}
