﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;
using SocketTest.Helpers;
using SocketTest.Models;
using System.IO;

namespace SocketTest.Sockets
{
    class Server : IDisposable
    {

        internal event Action<Peer_Model> Connected;
        internal event Action<Peer_Model> Disconnected;

        internal List<Peer_Model> Client_list = new List<Peer_Model>();
        Socket tcpServer;

        public Server()
        {
            InitServer();
        }

        public void InitServer()
        {
            UnInitServer();

            tcpServer = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            tcpServer.Bind(new IPEndPoint(IPAddress.Any, 4500));
            tcpServer.UseOnlyOverlappedIO = true;
            SocketHelper.SetKeepAlive(tcpServer);
            tcpServer.LingerState = new LingerOption(true, 0);
            tcpServer.Listen(50);

            var accept = new SocketAsyncEventArgs();
            accept.Completed += new EventHandler<SocketAsyncEventArgs>(Accept_Completed);

            tcpServer.AcceptAsync(accept);
        }

        private void Accept_Completed(object sender, SocketAsyncEventArgs e)
        {

            if (e.SocketError == SocketError.Success)
            {
                var sock = e.AcceptSocket;

                sock.UseOnlyOverlappedIO = true;
                //Set_KeepAlive(sock);

                var new_peer = new Peer_Model()
                {
                    peer = sock,
                    IP = ((IPEndPoint)sock.RemoteEndPoint).Address.ToString(),
                    Port = ((IPEndPoint)(sock.RemoteEndPoint)).Port
                };

                lock (Client_list) Client_list.Add(new_peer);

                Connected?.Invoke(new_peer);

                var receive = new SocketAsyncEventArgs();
                var buf = new byte[1];
                receive.SetBuffer(buf, 0, buf.Length);
                receive.UserToken = new_peer;
                receive.Completed += Receive_Completed;

                sock.ReceiveAsync(receive);

                e.AcceptSocket = null;
                tcpServer.AcceptAsync(e);

            }
            else if (e.SocketError == SocketError.OperationAborted) { }
            else
            {
                e.AcceptSocket = null;
                tcpServer.AcceptAsync(e);
            }
        }

        private void Receive_Completed(object sender, SocketAsyncEventArgs e)
        {
            var peerInfo = e.UserToken as Peer_Model;
            var sock = peerInfo.peer;

            if (e.SocketError == SocketError.Success && sock.IsConnected() && e.BytesTransferred == 1)
            {
                sock.ReceiveAsync(e);
            }
            else if (e.SocketError == SocketError.OperationAborted) { }
            else
                Delete_Peer(peerInfo);
        }

        public void Delete_Peer(Peer_Model _peer)
        {
            lock (Client_list)
                Client_list.Remove(_peer);

            _peer.Dispose();
            Disconnected?.Invoke(_peer);
            _peer = null;
        }

        void UnInitServer()
        {

            tcpServer?.Close();
            tcpServer?.Dispose();
            tcpServer = null;

            while (Client_list.Count > 0)
            {
                Peer_Model peer = Client_list[0];
                peer.Dispose();

                Client_list.Remove(peer);
            }

            Client_list.Clear();
        }

        #region IDisposable Support

        private bool disposedValue;

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    UnInitServer();
                    // TODO: 관리형 상태(관리형 개체)를 삭제합니다.
                }

                // TODO: 비관리형 리소스(비관리형 개체)를 해제하고 종료자를 재정의합니다.
                // TODO: 큰 필드를 null로 설정합니다.
                disposedValue = true;
            }
        }

        // // TODO: 비관리형 리소스를 해제하는 코드가 'Dispose(bool disposing)'에 포함된 경우에만 종료자를 재정의합니다.
        // ~Server()
        // {
        //     // 이 코드를 변경하지 마세요. 'Dispose(bool disposing)' 메서드에 정리 코드를 입력합니다.
        //     Dispose(disposing: false);
        // }

        public void Dispose()
        {
            // 이 코드를 변경하지 마세요. 'Dispose(bool disposing)' 메서드에 정리 코드를 입력합니다.
            Dispose(disposing: true);
            GC.SuppressFinalize(this);
        }

        #endregion
    }
}
